@@ -10,13 +10,20 @@
 #include <NTL/ctools.h>
 #include <NTL/new.h>
 
+#include <utility>
 #include <iostream>
 #include <new>
 #include <stdexcept>
 
 #include <cstdlib>
 #include <cmath>
+#include <cstring>
 
+#ifdef NTL_SAFE_VECTORS
+
+#include <type_traits>
+
+#endif
 
 
 #if (defined(NTL_THREADS) && defined(__GNUC__) && !defined(NTL_DISABLE_TLS_HACK))
@@ -35,6 +42,23 @@
 #define NTL_SNS std ::
 #define NTL_USE_SNS using namespace std;
 
+#define NTL_IMPORT_FROM_STD \
+   using NTL_SNS abs; \
+   using NTL_SNS ceil; \
+   using NTL_SNS exp; \
+   using NTL_SNS fabs; \
+   using NTL_SNS floor; \
+   using NTL_SNS ldexp; \
+   using NTL_SNS log; \
+   using NTL_SNS sqrt; \
+   using NTL_SNS ostream;  \
+   using NTL_SNS istream;  \
+   using NTL_SNS cerr;  \
+   using NTL_SNS ifstream;  \
+   using NTL_SNS ofstream; 
+
+
+
 #ifndef NTL_LEGACY_NO_NAMESPACE
 
 // This wraps NTL in the NTL namespace.
@@ -50,15 +74,8 @@
 // functions that are both overloaded by NTL and are used in
 // the implementation of NTL.
 
-#define NTL_START_IMPL NTL_USE_SNS NTL_OPEN_NNS \
-   using NTL_SNS abs; \
-   using NTL_SNS ceil; \
-   using NTL_SNS exp; \
-   using NTL_SNS fabs; \
-   using NTL_SNS floor; \
-   using NTL_SNS ldexp; \
-   using NTL_SNS log; \
-   using NTL_SNS sqrt; 
+#define NTL_START_IMPL NTL_OPEN_NNS NTL_IMPORT_FROM_STD
+
 
 #define NTL_END_IMPL NTL_CLOSE_NNS
 
@@ -73,7 +90,7 @@
 #define NTL_USE_NNS 
 #define NTL_NNS 
 
-#define NTL_START_IMPL
+#define NTL_START_IMPL NTL_IMPORT_FROM_STD
 #define NTL_END_IMPL
 
 #endif
@@ -471,7 +488,7 @@ public:
    WrappedPtr() : rep(0) { }
    void operator=(const raw_ptr& _rep)  { rep = _rep; }
 
-   ~WrappedPtr() { Deleter::apply(rep); } 
+   ~WrappedPtr() { if (rep) Deleter::apply(rep); } 
 
    operator const raw_ptr& () const { return rep; }
    operator raw_ptr& () { return rep; }
@@ -479,10 +496,17 @@ public:
    const raw_ptr* operator&() const { return &rep; }
    raw_ptr* operator&() { return &rep; }
 
-   void kill() { Deleter::apply(rep); rep = 0; }
+   void kill() { if (rep) { Deleter::apply(rep); rep = 0; } }
 
    void swap(WrappedPtr& other) { _ntl_swap(rep, other.rep); }
 
+   void move(WrappedPtr& other) 
+   {
+      WrappedPtr tmp;
+      tmp.swap(other);
+      tmp.swap(*this);
+   }
+
 };
 
 template<class T, class Deleter>
@@ -614,8 +638,12 @@ operator+(scope_guard_builder b, F&& f)
 
 
 class DummyScopeGuard {
+  bool active;
 public:
-   void relax() { }
+   DummyScopeGuard() : active(true) { }
+   ~DummyScopeGuard() { if (active) TerminalError("unexpected exception"); }
+   
+   void relax() { active = false; }
 };
 
 #define NTL_SCOPE(var) DummyScopeGuard var; if (false)
@@ -743,7 +771,7 @@ struct ll_type {
 };
 
 
-static inline void 
+inline void 
 ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
 {
   unsigned long hi, lo;
@@ -757,7 +785,7 @@ ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
    );
 }
 
-static inline void 
+inline void 
 ll_imul_add(ll_type& x, unsigned long a, unsigned long b)
 {
   unsigned long hi, lo;
@@ -771,7 +799,7 @@ ll_imul_add(ll_type& x, unsigned long a, unsigned long b)
    );
 }
 
-static inline void 
+inline void 
 ll_mul(ll_type& x, unsigned long a, unsigned long b)
 {
    __asm__ (
@@ -782,7 +810,7 @@ ll_mul(ll_type& x, unsigned long a, unsigned long b)
    );
 }
 
-static inline void 
+inline void 
 ll_imul(ll_type& x, unsigned long a, unsigned long b)
 {
    __asm__ (
@@ -793,7 +821,7 @@ ll_imul(ll_type& x, unsigned long a, unsigned long b)
    );
 }
 
-static inline void
+inline void
 ll_add(ll_type& x, unsigned long a)
 {
    __asm__ (
@@ -819,8 +847,30 @@ ll_add(ll_type& x, unsigned long a)
 
 // NOTE: this implementation assumes that shamt is in the range 
 // 0..NTL_BITS_PER_LONG-1
+
+#if 1
+
+// The shrd instruction can be very slow on some
+// machines.  Two shifts is usually just as good.
+
 template<long shamt>
-static inline unsigned long
+unsigned long
+ll_rshift_get_lo(ll_type x)
+{
+   unsigned long res;
+   if (shamt) 
+      res = (x.lo >> shamt) | (x.hi << (NTL_BITS_PER_LONG-shamt));
+   else
+      res = x.lo;
+      
+   return res;
+}
+
+#else
+
+
+template<long shamt>
+unsigned long
 ll_rshift_get_lo(ll_type x)
 {
    if (shamt) {
@@ -834,21 +884,23 @@ ll_rshift_get_lo(ll_type x)
    return x.lo;
 }
 
+#endif
 
-static inline unsigned long 
+
+inline unsigned long 
 ll_get_lo(const ll_type& x)
 {
    return x.lo;
 }
 
-static inline unsigned long 
+inline unsigned long 
 ll_get_hi(const ll_type& x)
 {
    return x.hi;
 }
 
 
-static inline void
+inline void
 ll_init(ll_type& x, unsigned long a)
 {
    x.lo = a;
@@ -863,7 +915,7 @@ typedef NTL_ULL_TYPE ll_type;
 // NOTE: the following functions definitions should serve as
 // documentation, as well.
 
-static inline void 
+inline void 
 ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
 {
    x += ((ll_type) a)*((ll_type) b);
@@ -871,12 +923,12 @@ ll_mul_add(ll_type& x, unsigned long a, unsigned long b)
 
 // a and b should be representable as positive long's,
 // to allow for the most flexible implementation
-static inline void 
+inline void 
 ll_imul_add(ll_type& x, unsigned long a, unsigned long b)
 {
    x += ((ll_type) long(a))*((ll_type) long(b));
 }
-static inline void 
+inline void 
 ll_mul(ll_type& x, unsigned long a, unsigned long b)
 {
    x = ((ll_type) a)*((ll_type) b);
@@ -884,13 +936,13 @@ ll_mul(ll_type& x, unsigned long a, unsigned long b)
 
 // a and b should be representable as positive long's,
 // to allow for the most flexible implementation
-static inline void 
+inline void 
 ll_imul(ll_type& x, unsigned long a, unsigned long b)
 {
    x = ((ll_type) long(a))*((ll_type) long(b));
 }
 
-static inline void
+inline void
 ll_add(ll_type& x, unsigned long a)
 {
    x += a;
@@ -898,26 +950,26 @@ ll_add(ll_type& x, unsigned long a)
 
 // NOTE: shamt must be in the range 0..NTL_BITS_PER_LONG-1
 template<long shamt>
-static inline unsigned long
+unsigned long
 ll_rshift_get_lo(const ll_type& x)
 {
    return ((unsigned long) (x >> shamt));
 }
 
-static inline unsigned long 
+inline unsigned long 
 ll_get_lo(const ll_type& x)
 {
    return ((unsigned long) x);
 }
 
-static inline unsigned long 
+inline unsigned long 
 ll_get_hi(const ll_type& x)
 {
    return ((unsigned long) (x >> NTL_BITS_PER_LONG));
 }
 
 
-static inline void
+inline void
 ll_init(ll_type& x, unsigned long a)
 {
    x = a;
@@ -927,7 +979,7 @@ ll_init(ll_type& x, unsigned long a)
 #endif
 
 
-static inline unsigned long 
+inline unsigned long 
 ll_mul_hi(unsigned long a, unsigned long b)
 {
    ll_type x;
@@ -942,6 +994,134 @@ ll_mul_hi(unsigned long a, unsigned long b)
 
 
 
+#ifdef NTL_SAFE_VECTORS
+
+
+#define NTL_RELOC_TAG (relocatable)
+
+#define NTL_DECLARE_RELOCATABLE_WHEN(x) \
+constexpr bool DeclareRelocatableType x
+
+#if (defined(NTL_HAVE_COPY_TRAITS1))
+
+
+// This strategy is used on compilers that fully support C++11 type traits.
+// For some reason, is_trivially_copyable says "true" even if the class
+// has deleted it's copy constructor.  Which means it is not copyable at all.
+// So I added an explicit test for is_copy_constructible.
+// Just to be on the safe side, I check for a trivial destructor.
+
+// This strategy is checked in the CheckCOPY_TRAITS1.cpp program.
+
+
+template<class T>
+constexpr bool Relocate_aux_has_trivial_copy(T*)
+{
+   return  std::is_trivially_copyable<T>::value &&
+           std::is_trivially_destructible<T>::value &&
+           std::is_copy_constructible<T>::value;
+}
+
+#elif (defined(NTL_HAVE_COPY_TRAITS2))
+
+// This strategy is needed on GCC before v5.0, as the required type
+// traits are not impplemented.  Note that on a class with it's copy
+// constructors deleted, __has_trivial_copy is false on GCC before 4.9
+// and true startig with 4.9.  So I had to make use of SFINAE techniques
+// to make sure there actually is a non-deleted copy constructor.
+// Just to be on the safe side, I check for a trivial destructor.
+
+// This strategy is checked in the CheckCOPY_TRAITS1.cpp program.
+
+template <bool statement, typename out>
+struct Relocate_aux_Failable
+{
+     typedef out Type;
+};
+
+struct Relocate_aux_TwoChars { char d[2]; };
+
+template <typename T>
+struct Relocate_aux_has_copy
+{
+
+     static const T *MakeT();
+
+     template <typename U> // U and T are the same type
+     static typename Relocate_aux_Failable<(bool(sizeof U(*MakeT()))), char>::Type copy(int);
+
+     template <typename U>
+     static typename Relocate_aux_Failable<true, Relocate_aux_TwoChars>::Type copy(...);
+
+     enum { value =  sizeof( copy<T>(0) )  == 1 };
+};
+
+
+template<class T>
+constexpr bool Relocate_aux_has_trivial_copy(T*)
+{
+   return  __has_trivial_copy(T) &&
+           __has_trivial_destructor(T) &&
+           Relocate_aux_has_copy<T>::value;
+}
+
+
+#else
+
+#error "lacking compiler support for NTL_SAFE_VECTORS"
+
+#endif
+
+
+// NOTE: I've checked the correctness of the above strategies using
+// Godbolt's compiler explorer across a range of complilers
+// (clang, gcc, icc, MS).
+
+
+template<class T>
+constexpr bool DeclareRelocatableType(T*)
+{
+   return Relocate_aux_has_trivial_copy((T*)0);
+}
+ 
+
+#else
+
+#define NTL_RELOC_TAG (true)
+
+#define NTL_DECLARE_RELOCATABLE_WHEN(x) \
+inline bool DeclareRelocatableType x
+
+
+template<class T>
+inline bool DeclareRelocatableType(T*)
+{
+   return true;
+}
+
+#endif
+
+
+#define NTL_DECLARE_RELOCATABLE(x) NTL_DECLARE_RELOCATABLE_WHEN(x) \
+   { return true; }
+
+
+// Examples: 
+//   NTL_DECLARE_RELOCATABLE((int*))
+//   NTL_DECLARE_RELOCATABLE((Foo<int>*)) 
+//   template <class X, class Y> NTL_DECLARE_RELOCATABLE((Foo<X,Y>*))
+
+
+#if (NTL_CXX_STANDARD >= 2011)
+
+#define NTL_DEFAULT =default;
+
+#else
+
+#define NTL_DEFAULT {}
+
+#endif
+
 
 NTL_CLOSE_NNS
 
